(******************************************************************************)
(*  ocaml-fileutils: files and filenames common operations                    *)
(*                                                                            *)
(*  Copyright (C) 2003-2014, Sylvain Le Gall                                  *)
(*                                                                            *)
(*  This library is free software; you can redistribute it and/or modify it   *)
(*  under the terms of the GNU Lesser General Public License as published by  *)
(*  the Free Software Foundation; either version 2.1 of the License, or (at   *)
(*  your option) any later version, with the OCaml static compilation         *)
(*  exception.                                                                *)
(*                                                                            *)
(*  This library is distributed in the hope that it will be useful, but       *)
(*  WITHOUT ANY WARRANTY; without even the implied warranty of                *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file         *)
(*  COPYING for more details.                                                 *)
(*                                                                            *)
(*  You should have received a copy of the GNU Lesser General Public License  *)
(*  along with this library; if not, write to the Free Software Foundation,   *)
(*  Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA             *)
(******************************************************************************)

include FileUtilTypes
include FileUtilPermission
include FileUtilSize
include FileUtilSTAT
include FileUtilUMASK
include FileUtilLS
include FileUtilCHMOD
include FileUtilTEST
include FileUtilPWD
include FileUtilREADLINK
include FileUtilWHICH
include FileUtilMKDIR
include FileUtilTOUCH
include FileUtilFIND
include FileUtilRM
include FileUtilCP
include FileUtilMV
include FileUtilCMP
include FileUtilDU

type 'a error_handler = string -> 'a -> unit

module Mode = FileUtilMode
